package counter;

import com.google.common.base.Stopwatch;
import org.junit.FixMethodOrder;
import org.junit.Test;
import org.junit.runners.MethodSorters;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.TimeUnit;
import java.util.function.Supplier;

import static org.assertj.core.api.Assertions.assertThat;

@FixMethodOrder(MethodSorters.NAME_ASCENDING)
public class CounterTest {
    private final static int NR_OF_THREADS = 100;
    private final static int NR_OF_INCREMENTS = 10000;

    @Test
    public void testThreadUnsafeCounter() {
        final Counter counter = runWith(() -> new ThreadUnsafeCounter());
        assertThat(counter.getValue()).isNotEqualTo(NR_OF_INCREMENTS * NR_OF_THREADS);
    }

    @Test
    public void testThreadSafeCounter() {
        final Counter counter = runWith(() -> new ThreadSafe1Counter());
        assertThat(counter.getValue()).isEqualTo(NR_OF_INCREMENTS * NR_OF_THREADS);
    }

    @Test
    public void testThreadSafe2Counter() {
        final Counter counter = runWith(() -> new ThreadSafe2Counter());
        assertThat(counter.getValue()).isEqualTo(NR_OF_INCREMENTS * NR_OF_THREADS);
    }

    @Test
    public void testAtomicIntCounter() {
        final Counter counter = runWith(() -> new AtomicIntCounter());
        assertThat(counter.getValue()).isEqualTo(NR_OF_INCREMENTS * NR_OF_THREADS);
    }

    @Test
    public void testAtomicLongCounter() {
        Counter counter = runWith(() -> new AtomicLongCounter());
        assertThat(counter.getValue()).isEqualTo(NR_OF_INCREMENTS * NR_OF_THREADS);
    }

    private Counter runWith(Supplier<Counter> f) {

        doIt(f.get());
        Stopwatch sw = Stopwatch.createStarted();
        Counter counter = f.get();
        doIt(counter);
        sw.stop();
        System.out.println(counter + " expected: " + 2 * NR_OF_INCREMENTS * NR_OF_THREADS + " got: " + counter.getValue() + " in: " + sw.elapsed(TimeUnit.MILLISECONDS) + "ms");
        return counter;
    }

    private void doIt(Counter counter) {
        List<Thread> threads = new ArrayList<>();
        for (int i = 0; i < NR_OF_THREADS; ++i) {
            threads.add(new CountingThread(counter));
        }

        for (Thread t : threads) {
            t.start();
        }
        for (Thread t : threads) {
            try {
                t.join();
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        }
    }

    private static class CountingThread extends Thread {
        private final Counter counter;

        CountingThread(Counter counter) {
            this.counter = counter;
        }

        @Override
        public void run() {
            for (int i = 0; i < NR_OF_INCREMENTS; ++i) {
                counter.increment();
            }
        }
    }

}